--==============================================================================
--miniSipServer calling card service script
--
--Copyright (C) MYVOIPAPP,Inc. All Rights Reserved.
--
--author: Hong
--
--History:
--2011-03-05 Migrated from MSS core.  
--2011-03-12 Support full features.
--2011-03-18 Fix a bug: if the call is not answered, it is unnecessary to set CDR to MySQL database.
--2011-04-11 If MSS cannot get the prepaid rate for the destination, MSS should process it as "unavailable destination" and prompt fail tone to caller.
--2011-08-30 Support one-time fee in prepaid rate
--2016-08-05 Migrated from Python script
--==============================================================================

require "services"
require "utils.dba"
require "utils.date_time"

-- service states
local ECC_IDLE              = 0
local ECC_WAIT_CARD         = 1
local ECC_WAIT_PWD          = 2
local ECC_WAIT_DEST_NBR     = 3
local ECC_PLAY_DURATION     = 4
local ECC_PLAY_INVALID_CARD = 5
local ECC_PLAY_BALANCE      = 6
local ECC_PLAY_DEST_UNAVAIL = 7
local ECC_PLAY_SELECTION    = 8
local ECC_PLAY_BYE          = 9
local ECC_WAIT_ANSWER       = 10
local ECC_TALK              = 11

-- Timer
local CC_TIMER_PLAY     = 115 --Seconds
local CC_TIMER_WAIT_ANS = 120 --Seconds

--default values
local CC_CALLER_LEG = 1
local CC_CALLED_LEG = 2

local CC_FIRST_DIGIT_TIMEOUT = 60 -- Seconds
local CC_INTER_DIGIT_TIMEOUT = 6  -- Seconds

local CC_MAX_DURATION_FOR_FREE_CALL = 28800 -- Seconds

local CC_MAX_CARD_LEN = 32
local CC_MAX_PWD_LEN  = 32

local CC_DEFAULT_MATCH_TAG = "*"
local CC_DEFAULT_DIAL_PLAN = "default"

local CC_MAX_INPUT_CARD_CNT  = 3
local CC_MAX_TRY_DESTINATION = 3


--*************************************
--
-- Common functions
--
--*************************************
local function initCallInfo()
    fsm.dialPlan = fsm.origDialPlan
    fsm.allowDuration = 0
end

local function initCardInfo()
    fsm.card=""
    fsm.pwd=""
    fsm.balance=0
    fsm.rateId = 0
    
    fsm.rateInfo={}
    fsm.rateInfo.interval=0
    fsm.rateInfo.tariff=0
    fsm.rateInfo.onetimeFee=0
    
    fsm.dialPlanInd=0
    fsm.cardLockerInd = MSS_NO
end

local function initCDRInfo()
    fsm.setCDRInd   = MSS_NO
    
    fsm.setupTime   = 0
    fsm.answerTime  = 0
    fsm.releaseTime = 0
    fsm.cdrDuration = 0
    fsm.cdrFee      = 0
end

local function initAllParams()
    fsm.ivrInd = MSS_NO
    fsm.makeCallCnt = 0
    fsm.inputCardCnt = 0
    fsm.destNbr= ""
    
    initCallInfo()
    initCardInfo()
    initCDRInfo()
end

local function getLockerName()    
    return string.format("cc_card=%s", fsm.card)    
end

local function isCardLocked()
    if string.len(fsm.card)==0 then return MSS_NO end

    local lockerName = getLockerName()
    return hasLocker(lockerName)
end

local function lockCard()
    if fsm.cardLockerInd==MSS_YES or string.len(fsm.card)==0 then return end

    local lockerName = getLockerName()
    setLocker(lockerName)
    fsm.cardLockerInd = MSS_YES
    TraceInfo("\t lockCard")
end

local function unlockCard()
    if fsm.cardLockerInd==MSS_NO or string.len(fsm.card)==0 then return end

    local lockerName = getLockerName()
    clrLocker(lockerName)

    fsm.card=""
    fsm.cardLockerInd = MSS_NO
    TraceInfo("\t unlockCard")
end

local function sendRRBEs()
    SendRRBE(CC_CALLED_LEG, DP_ROUTE_SELECT_FAILURE,  DP_MON_INTERUPT)
    SendRRBE(CC_CALLED_LEG, DP_O_BUSY,                DP_MON_INTERUPT)
    SendRRBE(CC_CALLED_LEG, DP_O_NO_ANSWER,           DP_MON_INTERUPT)
    SendRRBE(CC_CALLED_LEG, DP_O_DISCONNECT,          DP_MON_INTERUPT)
    SendRRBE(CC_CALLED_LEG, DP_O_ANSWER,              DP_MON_INTERUPT)
end

local function connectDestNbr()
    TraceInfo("\t connectDestNbr")

     local connArg = CreateConnectArg()
     connArg.destNbr = fsm.destNbr
     connArg.dialPlanInd = MSS_YES
     connArg.dialPlan = fsm.dialPlan
     connArg.servInfoInd=MSS_YES
     connArg.suppressCallLevel = SERVINFO_SUPPRESS_CALL_LEVEL_YES

     SendConnectWithArg(connArg)
end

--*************************************
--
-- IVR functions
--
--*************************************
local function prepareIVR()
    TraceInfo("\t prepareIVR")
    if fsm.ivrInd==MSS_NO then 
        SendCTR(CC_CALLER_LEG)
        fsm.ivrInd = MSS_YES
    end
end

local function stopIVR()
    TraceInfo("\t stopIVR")
    if fsm.ivrInd==MSS_YES then
        SendDFC(CC_CALLER_LEG)
        fsm.ivrInd = MSS_NO
    end
end

local function requestCard()
    TraceInfo("\t requestCard")
    SendPNC(CC_CALLER_LEG, 0x84080001, CC_FIRST_DIGIT_TIMEOUT, CC_INTER_DIGIT_TIMEOUT)
end

local function requestPassword()
    TraceInfo("\t requestPassword")
    SendPNC(CC_CALLER_LEG, 0x84080002, CC_FIRST_DIGIT_TIMEOUT, CC_INTER_DIGIT_TIMEOUT)
end

local function requestDestNbr()
    TraceInfo("\t requestDestNbr")
    SendPNC(CC_CALLER_LEG, 0x84080003, CC_FIRST_DIGIT_TIMEOUT, CC_INTER_DIGIT_TIMEOUT)
end

local function requestSelection()
    local pnc = CreatePNC()
    pnc.infoSend_annID_annID = 0x84080009
    pnc.infoSend_numOfRepeat = 1
    pnc.maxNumOfDigits = 1

    SendPNCWithArg(CC_CALLER_LEG, pnc)
end

local function playBalance()
    TraceInfo("\t playBalance")
    local paArg = CreatePA()
    paArg.infoSend_annID_annID = 0x84080005
    paArg.infoSend_numOfRepeat = 1
    paArg.infoSend_duration = CC_TIMER_WAIT_ANS

    paArg.infoSend_varType = ESRF_INFO_VAR_PRICE
    paArg.infoSend_var_price = fsm.balance
    SendPA(CC_CALLER_LEG, paArg)
end

local function playDuration()
    TraceInfo("\t playDuration")
    local paArg = CreatePA()
    paArg.infoSend_annID_annID = 0x84080004
    paArg.infoSend_numOfRepeat = 1

    paArg.infoSend_varType = ESRF_INFO_VAR_DURATION
    paArg.infoSend_var_duration = fsm.allowDuration
    SendPA(CC_CALLER_LEG, paArg)
end

local function playBalanceAndRequestDest()
    TraceInfo("\t playBalanceAndRequestDest")
    local pncArg = CreatePNC()
    pncArg.infoSend_annID_annID = 0x84080005
    pncArg.infoSend_numOfRepeat = 1
    pncArg.infoSend_duration = CC_TIMER_WAIT_ANS
    pncArg.infoSend_varType = ESRF_INFO_VAR_PRICE
    pncArg.infoSend_var_price = fsm.balance
    pncArg.infoSend_varType2 = ESRF_INFO_VAR_MULTI_ID
    pncArg.infoSend_var_annID2 = 0x84080003

    SendPNCWithArg(CC_CALLER_LEG, pncArg)
end

local function playDestFail()
    TraceInfo("\t playDestFail")
    local paArg = CreatePA()
    paArg.infoSend_annID_annID = 0x84080006
    paArg.infoSend_numOfRepeat = 1
    paArg.infoSend_duration = CC_TIMER_WAIT_ANS

    SendPA(CC_CALLER_LEG, paArg)
end

local function playInvalidCard()
    TraceInfo("\t playInvalidCard")
    local paArg = CreatePA()
    paArg.infoSend_annID_annID = 0x84080007
    paArg.infoSend_numOfRepeat = 1
    paArg.infoSend_duration = CC_TIMER_WAIT_ANS

    SendPA(CC_CALLER_LEG, paArg)
end

local function playBye()
    TraceInfo("\t playBye")
    local paArg=CreatePA()
    paArg.infoSend_annID_annID = 0x84080008
    paArg.infoSend_numOfRepeat = 1
    paArg.infoSend_duration = CC_TIMER_WAIT_ANS

    SendPA(CC_CALLER_LEG, paArg)
end

local function playWaitMusic()
    TraceInfo("\t playWaitMusic")
    SendRingBackToneToCaller()
end

--*************************************
--
-- DB functions
--
--*************************************
local function dbHasDialPlanInd(option)
    return option&1
end


local function dbGetCardInfo(card)
    local sqlStr = string.format("select cPassword,nBalance,cDialPlan, nOption from tbl_callingcard_balance"..
                                 " where cCard='%s'", card)
    DbPrepare()
    local result = DbQuery(sqlStr)
    if MSS_ERR_OK~=result or 0==DbGetRecordCnts() then 
        DbEnd()
        return -1
    end

    DbFetchRow()
    
    fsm.pwd = DbGetRowStrVal(0)
    fsm.balance = DbGetRowIntVal(1)
    local dialPlan = DbGetRowStrVal(2)
    local option = DbGetRowIntVal(3)
    if dbHasDialPlanInd(option)>0 then
        fsm.dialPlan = dialPlan
        fsm.dialPlanInd = MSS_YES 
    end
    TraceInfo("\t\t balance="..fsm.balance..",password="..fsm.pwd)
    DbEnd()
    return 0
end

local function dbGetBalance()
    if string.len(fsm.card)==0 then return 0 end

    local sqlStr = string.format("select nBalance from tbl_callingcard_balance where"..
                                 " cCard='%s'", fsm.card)
    DbPrepare()
    local result = DbQuery(sqlStr)
    if MSS_ERR_OK~=result or 0==DbGetRecordCnts() then 
        DbEnd()
        return -1
    end
    DbFetchRow()
    local balance = DbGetRowIntVal(0)
    DbEnd()
    TraceInfo("\t\t balance="..balance)
    return balance                                 
end

local function dbSetBalance(new_balance)
    if string.len(fsm.card)==0 then return end

    local sqlStr = string.format("update tbl_callingcard_balance set nBalance='%d' "..
                                 "where cCard='%s'", new_balance, fsm.card)
    DbPrepare()
    DbQuery(sqlStr)
    DbEnd()
end

local function dbFeeMatrixGet(dial_plan, called_nbr)
    local sqlStr=string.format("select nRateId from tbl_callingcard_fee_matrix "..
                               "where cDialPlan='%s' and cRouteNbr='%s'",
                               dial_plan, called_nbr)
    DbPrepare()
    local result = DbQuery(sqlStr)
    if MSS_ERR_OK~=result or DbGetRecordCnts()<=0 then 
        DbEnd()
        return -1 
    end
    
    DbFetchRow()
    local val = DbGetRowIntVal(0)
    DbEnd()
    return val
end

local function dbFeeMatrixMatchBase(dial_plan, called_nbr)
    local sqlStr=string.format("select nRateId from tbl_callingcard_fee_matrix where "..
                        "cDialPlan='%s' and position(cRouteNbr in '%s')=1 "..
                        "order by length(cRouteNbr) desc limit 1",
                        dial_plan, called_nbr)
    DbPrepare()
    local result = DbQuery(sqlStr)
    if MSS_ERR_OK~=result or DbGetRecordCnts()<=0 then 
        DbEnd()
        return -1 
    end
    
    DbFetchRow()
    local val = DbGetRowIntVal(0)
    DbEnd()
    return val
end

local function dbFeeMatrixMatch(dial_plan, called_nbr)
    local result = dbFeeMatrixMatchBase(dial_plan, called_nbr)
    if result>0 then return result end

    result = dbFeeMatrixGet(dial_plan, CC_DEFAULT_MATCH_TAG)
    if result>0 then return result end

    if dial_plan==CC_DEFAULT_DIAL_PLAN then return -1 end

    return dbFeeMatrixMatch(CC_DEFAULT_DIAL_PLAN, called_nbr)
end

local function dbGetRateInfo()
    if fsm.rateId<=0 then return -1 end

    local sqlStr=string.format("select tariff_interval, tariff, onetime_fee from tbl_prepaid_rate "..
                               "where rate_id=%d", fsm.rateId)
    DbPrepare()
    local result = DbQuery(sqlStr)
    if MSS_ERR_OK~=result or DbGetRecordCnts()<=0 then 
        DbEnd()
        return -2 
    end

    DbFetchRow()
    fsm.rateInfo.interval = DbGetRowIntVal(0)
    fsm.rateInfo.tariff = DbGetRowIntVal(1)
    fsm.rateInfo.onetimeFee = DbGetRowIntVal(2)
    DbEnd()

    TraceInfo("\t\t rate info: interval="..fsm.rateInfo.interval..
             ",tariff="..fsm.rateInfo.tariff..
             ",onetimeFee="..fsm.rateInfo.onetimeFee)
    return 1
end

local function dbSetCDR()
    if string.len(fsm.card)==0 or fsm.answerTime==0 or fsm.setCDRInd==MSS_YES then return end

    TraceInfo("\t dbSetCDR")
    local sqlStr=string.format("insert into tbl_callingcard_cdr (cCard, cCalled, StartTime, ConnTime, ReleaseTime, nDuration, nFee) "..
        "values('%s', '%s', '%s', '%s', '%s', %d, %d)",
        fsm.card, fsm.destNbr,timeToStr(fsm.setupTime), timeToStr(fsm.answerTime), timeToStr(fsm.releaseTime), fsm.cdrDuration, fsm.cdrFee)
    DbPrepare()
    DbQuery(sqlStr)
    DbEnd()

    initCDRInfo()
    fsm.setCDRInd = MSS_YES
end

--*************************************
--
-- calling card functions
--
--*************************************
local function calculateDuration()
    TraceInfo("\t calculateDuration")
    
    fsm.rateId = dbFeeMatrixMatch(fsm.dialPlan, fsm.destNbr)
    if fsm.rateId<0 then
        TraceInfo("\t\t cannot get rate ID.")
        return -1 
    end

    local result = dbGetRateInfo()
    if result<0 then
        TraceInfo("\t\t cannot get rate info")
        return -2
    end

    if 0==fsm.rateInfo.tariff then 
        fsm.allowDuration = CC_MAX_DURATION_FOR_FREE_CALL 
    else
        local balance = fsm.balance - fsm.rateInfo.onetimeFee
        if balance < fsm.rateInfo.tariff then
            TraceInfo("\t\t balance is not enough")
            return -3
        end

        result = ToInt(balance/fsm.rateInfo.tariff)
        fsm.allowDuration = result * fsm.rateInfo.interval    
    end
    TraceInfo("\t\t allow duration: "..fsm.allowDuration)
    return 1
end

local function calculateCdrFee()
    TraceInfo("\t calculateCdrFee")
    if fsm.answerTime==0 then
        fsm.cdrFee = 0
        fsm.cdrDuration = 0
        return
    end

    if fsm.releaseTime>fsm.answerTime then
        fsm.cdrDuration = ToInt(Duration(fsm.releaseTime, fsm.answerTime))    
    end
    TraceInfo("\t\t cdr duration="..fsm.cdrDuration)

    if fsm.rateInfo.tariff==0 then 
        fsm.cdrFee=0
    elseif fsm.cdrDuration>=fsm.allowDuration then
        fsm.cdrFee = fsm.balance
    else
        local val = ToInt(fsm.cdrDuration/fsm.rateInfo.interval)
        fsm.cdrFee = val * fsm.rateInfo.tariff

        -- even it is less than an interval, treat it as an interval
        val = ToInt(fsm.cdrDuration%fsm.rateInfo.interval)
        if val>0 then
            fsm.cdrFee = fsm.cdrFee + fsm.rateInfo.tariff
        end 

        -- add one-time fee
        if fsm.cdrDuration>0 then
            fsm.cdrFee = fsm.cdrFee + fsm.rateInfo.onetimeFee
        end

        if fsm.cdrFee>fsm.balance then 
            fsm.cdrFee = fsm.balance
        end
    end

    TraceInfo("\t\t cdr fee="..fsm.cdrFee)
end

local function updateBalance()
    TraceInfo("\t updateBalance")
    if fsm.cdrFee==0 then return end

    local balance = dbGetBalance()
    if balance < fsm.cdrFee then
        balance = 0
    else
        balance = balance - fsm.cdrFee
    end

    dbSetBalance(balance)
end

--*************************************
--
-- Procedures for messages and states
--
--*************************************
local function recvCard()
    TraceInfo("\t recvCard")
    fsm.card = currMessage.digits
    TraceInfo("\t\t card is "..fsm.card)
    fsm.inputCardCnt = fsm.inputCardCnt + 1

    local result = dbGetCardInfo(fsm.card)
    if result>=0 then
        if isCardLocked()==MSS_YES then
            TraceInfo("\t\t current card is used by another user.")
            result = -100
        end
    end

    if result<0 then
        TraceInfo("\t\t fail to get card information. result="..result)
        playInvalidCard()
        EnterState(ECC_PLAY_INVALID_CARD)
        return
    end

    -- lock card
    lockCard()

    -- check password
    if string.len(fsm.pwd)==0 then
        playBalance()
        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_PLAY_BALANCE)
    else
        requestPassword()
        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_WAIT_PWD)
    end    
end

local function recvPWD()
    TraceInfo("\t recvPWD")
    local pwd = currMessage.digits
    TraceInfo("\t\t user inputs password: "..pwd)

    -- check password
    if string.len(pwd)==0 or pwd~=fsm.pwd then
        unlockCard()
        playInvalidCard()
        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_PLAY_INVALID_CARD)
        return 
    end

    playBalanceAndRequestDest()
    StartTimer(CC_TIMER_PLAY)
    EnterState(ECC_WAIT_DEST_NBR)
end

local function callFail()
    TraceInfo("\t callFail")

    stopIVR() -- stop waiting tone
    fsm.releaseTime = Now()
    prepareIVR() -- prepare IVR to play fail tone 
    playDestFail()

    StartTimer(CC_TIMER_PLAY)
    EnterState(ECC_PLAY_DEST_UNAVAIL)
end

local function callConnected()
    TraceInfo("\t callConnected")
    fsm.answerTime = Now()
    stopIVR()
    SendContinue()

    StartTimer(fsm.allowDuration)
    EnterState(ECC_TALK)
end

local function callEnd()
    TraceInfo("\t callEnd")
    fsm.releaseTime = Now()
    calculateCdrFee()
    updateBalance()
    dbSetCDR()

    prepareIVR()
    if fsm.makeCallCnt>CC_MAX_TRY_DESTINATION then
        playBye()
        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_PLAY_BYE)
    else
        requestSelection()
        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_PLAY_SELECTION)
    end
end

local function recvDestNbr()
    TraceInfo("\t recvDestNbr")
    fsm.destNbr = currMessage.digits
    TraceInfo("\t\t destination number: "..fsm.destNbr)
    if string.len(fsm.destNbr)==0 then EndService() end

    local result = calculateDuration()
    if result<0 then
        callFail()
        return
    end

    playDuration()
    StartTimer(CC_TIMER_WAIT_ANS)
    EnterState(ECC_PLAY_DURATION)
end

local function makeCall()
    TraceInfo("\t makeCall")
    fsm.makeCallCnt = fsm.makeCallCnt + 1
    initCDRInfo()
    sendRRBEs()
    playWaitMusic()
    connectDestNbr()

    fsm.setupTime = Now()
    StartTimer(CC_TIMER_WAIT_ANS)
    EnterState(ECC_WAIT_ANSWER)
end

local function dialNewDest()
    TraceInfo("\t dialNewDest")
    fsm.balance = dbGetBalance()
    initCallInfo()
    fsm.destNbr=""

    playBalance()
    StartTimer(CC_TIMER_PLAY)
    EnterState(ECC_PLAY_BALANCE)
end

local function dialSameDest()
    TraceInfo("\t dialSameDest")
    fsm.balance = dbGetBalance()
    initCallInfo()

    local result = calculateDuration()
    if result<0 then 
        callFail()
        return
    end

    playDuration()
    StartTimer(CC_TIMER_PLAY)
    EnterState(ECC_PLAY_DURATION) 
end

local function useNewCard()
    TraceInfo("\t useNewCard")

    fsm.makeCallCnt=0
    fsm.inputCardCnt=0
    fsm.destNbr=""

    unlockCard()
    initCallInfo()
    initCardInfo()
    initCDRInfo()

    requestCard()
    StartTimer(CC_TIMER_PLAY)
    EnterState(ECC_WAIT_CARD)
end

local function recvSelection()
    TraceInfo("\t recvSelection")
    local val = currMessage.digits
    TraceInfo("\t\t selection="..val)

    if val=="1" then -- dial new destination
        dialNewDest()
    elseif val=="2" then -- dial the same destination
        dialSameDest()
    elseif val=="3" then -- use new calling card
        useNewCard()
    else
        playBye()

        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_PLAY_BYE)
    end        
end

local function onIdleProc(currMsg)
    TraceInfo("callingCard:onIdleProc")
    
    initAllParams()    
    fsm.setupTime = Now()
    TraceInfo("\t current dial plan is "..fsm.dialPlan)
    
    prepareIVR()
    requestCard()
    
    StartTimer(CC_TIMER_WAIT_ANS)
    EnterState(ECC_WAIT_CARD)
end

local function onWaitCardProc(currMsg)
    TraceInfo("onWaitCardProc")
    if currMsg==ESCP_EVT_PNC_RESULT then
        recvCard()
    else
        EndService()
    end
end

local function onWaitPwdProc(currMsg)
    TraceInfo("onWaitPwdProc")
    if currMsg==ESCP_EVT_PNC_RESULT then
        recvPWD()
    else
        EndService()
    end
end

local function onWaitDestNbrProc(currMsg)
    TraceInfo("onWaitDestNbrProc")
    if currMsg==ESCP_EVT_PNC_RESULT then
        recvDestNbr()
    else
        EndService()
    end
end

local function onPlayBalanceProc(currMsg)
    TraceInfo("onPlayBalanceProc")
    if currMsg==ESCP_EVT_SRR then
        requestDestNbr()

        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_WAIT_DEST_NBR)
    else
        EndService()
    end
end

local function onPlayDurationProc(currMsg)
    TraceInfo("onPlayDurationProc")
    if currMsg==ESCP_EVT_SRR then   
        makeCall()
    else
        EndService()
    end
end

local function onWaitAnswerProc(currMsg)
    TraceInfo("onWaitAnswerProc")
    if currMsg==ESCP_EVT_ERB then
        if currMessage.edpEvent==DP_O_ANSWER then
            callConnected()
        else
            callFail() 
        end
    else
        EndService()
    end
end

local function onTalkProc(currMsg)
    TraceInfo("onTalkProc")
    if currMsg==ESCP_EVT_TIMEOUT then
        prepareIVR()
        playBye()

        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_PLAY_BYE)
    elseif currMsg==ESCP_EVT_ERB then -- only O_DISC2 should be received
        callEnd()
    else
        EndService()
    end
end

local function onPlaySelectionProc(currMsg)
    TraceInfo("onPlaySelectionProc")

    if currMsg==ESCP_EVT_PNC_RESULT then
        recvSelection()
    elseif currMsg==ESCP_EVT_SRF_ERR then
        playBye()
        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_PLAY_BYE)
    else
        EndService()
    end
end

local function onPlayDestUnavailProc(currMsg)
    TraceInfo("onPlayDestUnavailProc")

    if fsm.makeCallCnt>=CC_MAX_TRY_DESTINATION then
        playBye()
        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_PLAY_BYE)
    else
        playBalance()
        StartTimer(CC_TIMER_PLAY)
        EnterState(ECC_PLAY_BALANCE)
    end
end

local function onPlayInvalidCardProc(currMsg)
    TraceInfo("onPlayInvalidCardProc")

    if currMsg==ESCP_EVT_SRR then
        if fsm.inputCardCnt>=CC_MAX_INPUT_CARD_CNT then
            playBye()
            StartTimer(CC_TIMER_PLAY)
            EnterState(ECC_PLAY_BYE)
        else
            requestCard()
            StartTimer(CC_TIMER_PLAY)
            EnterState(ECC_WAIT_CARD)
        end
    else
        EndService()
    end
end

local function onPlayByeProc(currMsg)
    TraceInfo("onPlayByeProc")

    EndService()
end

--*************************************
--
-- Service main function
--
--*************************************
function MainProc(currMsg)
    StopTimer()
    
    if fsm.state==ECC_IDLE then
        onIdleProc(currMsg)

    elseif fsm.state==ECC_WAIT_CARD then
        onWaitCardProc(currMsg)

    elseif fsm.state==ECC_WAIT_PWD then
        onWaitPwdProc(currMsg)

    elseif fsm.state==ECC_PLAY_BALANCE then
        onPlayBalanceProc(currMsg)

    elseif fsm.state==ECC_WAIT_DEST_NBR then
        onWaitDestNbrProc(currMsg)

    elseif fsm.state==ECC_PLAY_DURATION then
        onPlayDurationProc(currMsg)

    elseif fsm.state==ECC_WAIT_ANSWER then
        onWaitAnswerProc(currMsg)

    elseif fsm.state==ECC_TALK then
        onTalkProc(currMsg)

    elseif fsm.state==ECC_PLAY_SELECTION then
        onPlaySelectionProc(currMsg)

    elseif fsm.state==ECC_PLAY_DEST_UNAVAIL then
        onPlayDestUnavailProc(currMsg)

    elseif fsm.state==ECC_PLAY_INVALID_CARD then
        onPlayInvalidCardProc(currMsg)
    
    elseif fsm.state==ECC_PLAY_BYE then
        onPlayByeProc(currMsg)

    else
        TraceInfo("\t unknown state: "..fsm.state)
        EndService()
    end    
end


function DestroyProc()
    TraceInfo("DestroyProc")
    fsm.releaseTime = Now()

    calculateCdrFee()
    updateBalance()
    dbSetCDR()
    SendReleaseCall()
    unlockCard()
end
